"""Main pymend module."""

from .pymend import PyComment, __author__, __copyright__, __licence__, __version__

name = "pymend"  # pylint: disable=invalid-name

__all__ = ["PyComment", "__version__", "__copyright__", "__author__", "__licence__"]
