"""_summary_."""
def func1(param1):
    """Function to test issue #49.

    Parameters
    ----------
    param1 : object
        Description of param1

    Returns
    -------
    None
        Description for return value

    Raises
    ------
    AssertionError
        Raises an error if run
    """
    assert False
