"""_summary_."""


def ex1(
    docstring: str,
    quotes: str = '"""',
) -> None:
    pass


def ex2(
    docstring: str,
    quotes: str = '"""',
) -> None:
    """_summary_.

    Parameters
    ----------
    docstring : str
        The input docstring
    quotes : str
        The quotes to use.
    """
    print(x)

def ex3(
    docstring: str,
    quotes: str = '"""',
) -> None:
    """_summary_.

    Parameters
    ----------
    docstring : str
        The input docstring
    quotes : str
        The quotes to use. (Default value = '\"\"\"')
    """
    print(x)
