"""_summary_."""

def test1(param: int):
    """Short summary.
    Parameters
    ----------
    param : int
        description.
    """
    pass

def test2(param: int):
    """Short summary.

    Long description
    Parameters
    ----------
    param : int
        description.
    """
    pass

def test3(param: int):
    """Short summary.

    """
    pass

def test4(param: int):
    """Short summary.

    Long description

    


    """
    pass
