"""_summary_."""

def raises():
    """_summary_.

    Raises
    ------
    ValueError
        _description_
    TypeError
        _description_
    """
    raise ValueError
    raise KeyError
    raise TypeError(msg)
    raise ValueError